# REST data-logger with Tornado and MongoDB with Docker

## Application overview
The presented application is an example of a RESTful-based data-logger to be deployed on an embedded device acting as an 
intelligent edge node in large connected networks. Through a developed Web service application based on the *Tornado web 
framework*, data can be easily inserted and retrieved from a *MongoDB* instance by remote devices.

In the tutorial the application is proposed as a **Docker**-based solution in which each component is a separate container. 
**Docker Compose** is used to simplify the build of custom Docker images, their deployment, and configuration. The application
 can be easily tested on personal computers and embedded devices supporting Docker and Docker Compose. The whole solution 
 has been successfully deployed on Variscite Development Kits both on Debian and Yocto distros.

## Prerequisites
The following components are necessary to test the application:
* Docker. Installation instructions in [1](#references);
* Docker Compose. Installation instructions in [2](#references).

## Building and running the application
Once the application has been downloaded all the components are organized according to the following tree: 

```
RestDataLogger
│   Readme.md
│   docker-compose.yml
│
└───RestApplication
│   │   Dockerfile
│   │
│   └───app
│       │   WebService.py
│       │   requirements.txt
│ 
└───Mongodb
│   │   mongo-init.js
│
└───nginx
    │   Dockerfile
    │
    └───conf.d
        │   app.conf
```
To build and run the application the following command must be executed in the RestDataLogger directory:
```
$ docker-compose up -d
```
According to the content of the docker-compose.yml file, all the components of the application are built and executed. Once 
the whole process ends, by executing the command
```
$ docker ps
```
all the running containers can be listed.  If everything ends correctly the following output should be visualized:
```
CONTAINER ID  IMAGE            COMMAND                  CREATED          STATUS          PORTS                 NAMES
117a1ec2b982  nginxapp         "nginx -g 'daemon of…"   24 seconds ago   Up 21 seconds   0.0.0.0:80->80/tcp,   nginxapp
                                                                                         0.0.0.0:443->443/tcp   
edbc7cb7df4e  restapplication  "python -u ./WebServ…"   27 seconds ago   Up 10 seconds   8890/tcp               restapp
25c65be9b754  mongo:4.2.5      "docker-entrypoint.s…"   31 seconds ago   Up 26 seconds   27017/tcp              mongodb
```
The application can be tested by inserting new data in the MongoDB through the *insertdata* API:
```
curl -i -X POST \
   -H "Content-Type:application/json" \
   -d \
'{
  "id" : "20",
  "value1" : "100",
  "value2" : "30",
  "TimestampAcquisition" : "2020-03-29 17:20:25"
}' \
 'http://localhost:80/restfuldatalogger/insertdata'
```
and by retrieving data through the *readdata* API:
```
curl -i -X GET \
   -H "Content-Type:application/json" \
 'http://localhost:80/restfuldatalogger/readdata/20'
```
If everything is working correctly both curl commands will return Success.

In case it is necessary to restart the application with a clean environment the following command can be executed before 
calling again the docker-compose command:
```
$ docker system prune --volumes
```

## Application description
In this section, all the components of the application are described by starting from the docker-compose.yml file.

### 1.Docker compose stack configuration
With docker compose the application infrastructure can be defined in a single file and built with a single command. In the 
docker-compose.yml file, the application infrastructure is defined as individual services. The services can be connected 
by networking parameters, while their data can be made persistent by defining mapping volumes.

From the docker-compose.yml structure, three main individual services can be identified. Moreover, general volumes and 
networks keywords are used:
```
version: '3.3'
services:

  tornado:
   .....

  mongodb:
   .....

  nginx:
   .....

  volumes:
   .....

  networks:
   .....

```

The *tornado* section is related to the developed Web service application based on the Tornado web framework:
```
  tornado:
    build:
      context: RestApplication
      dockerfile: Dockerfile
    container_name: restapp
    image: restapplication
    restart: unless-stopped
    environment:
      APP_PORT: 8890
      MONGODB_HOSTNAME: mongodb
      MONGODB_PORT: 27017
      MONGODB_DATABASE: restappdb
      MONGODB_USERNAME: user
      MONGODB_PASSWORD: userpassword
    volumes:
      - appdata:/usr/src/app/log
    depends_on:
      - mongodb
    networks:
      - frontend
      - backend
```
In the docker compose syntax it means that the image must be built by using the Dockerfile in the RestApplication directory, 
the new image name is tagged as restapplication. The container_name property defines the name of the container. The restart 
property defines how the container should be restarted in case any error occurs. The environment property contains the environment 
variables that are passed to the container. The volumes property defines the volumes the service is using. The networks property 
specifies frontend and backend as the networks the application will have access to. The depends_on property defines a service 
that is necessary to let the Web service application to function properly. In this case, the mongodb service, which acts 
as a database. 

The *mongodb* section is related to the MongoDB service used to store inserted data:
```
 mongodb:
    image: mongo:4.2.5
    container_name: mongodb
    restart: unless-stopped
    command: mongod --auth
    environment:
      MONGO_INITDB_ROOT_USERNAME: admin
      MONGO_INITDB_ROOT_PASSWORD: adminpassword
      MONGO_INITDB_DATABASE: restappdb
      MONGO_INITDB_USER_USERNAME: user
      MONGO_INITDB_USER_PASSWORD: userpassword
      MONGODB_DATA_DIR: /data/db
    volumes:
      - ./Mongodb/mongo-init.js:/docker-entrypoint-initdb.d/mongo-init.js:ro
      - mongodata:/data/db
    networks:
      - backend

```
In this case, a prebuilt docker image available into the Docker Hub is used. Moreover, some environment variables are passed 
to the container, as well as a database initialization script to create a custom db and user in the default image (first 
row in the volumes property). In this case, the networks property includes only the backend, since the database will not 
be accessible from the frontend.

The *nginx* section is related to the Nginx, used  as a reverse proxy to forward requests tornado application:
```
  nginx:
    build:
      context: nginx
      dockerfile: Dockerfile
    image: nginxapp
    container_name: nginxapp
    restart: unless-stopped
    environment:
      APP_SERVER_PORT: 8890
    ports:
      - "80:80"
    volumes:
      - nginxdata:/var/log/nginx
    depends_on:
      - tornado
    networks:
      - frontend
```
Like the tornado service, nginx is built through a Dockerfile, and then the container is configured with the environment 
and port properties. Networks property includes only the frontend since it is the nginx service the one delegated to accept 
REST requests from devices to forward them to the tornado service.

### 2.Dockerized Tornado Web service
The developed Web service application simply implements two REST methods, a POST method used to insert data, and a GET method
to retrieve data. In the WebService.py file both methods are defined with the following code:
```
    app = tornado.web.Application([
        (r"/health", HealthHandler),
        (r"/restfuldatalogger/insertdata", InsertDataHandler, {'db': db}),
        (r"/restfuldatalogger/readdata/([0-9]+)", ReadDataHandler, {'db': db})
    ])
```
where db is the MongoDB database created to store the data, and defined in the internal environment variables:
```
    MONGODB_HOSTNAME = os.environ.get("MONGODB_HOSTNAME", 'localhost')
    MONGODB_PORT = int(os.environ.get("MONGODB_PORT", 27017))
    MONGODB_DATABASE=os.environ.get("MONGODB_DATABASE", 'appdb')
    MONGODB_USERNAME=os.environ.get("MONGODB_USERNAME", '')
    MONGODB_PASSWORD=os.environ.get("MONGODB_PASSWORD", '')
```
The POST method accepts a json message in the input. A possible message  is reported in the following as a matter of example:
```
{
  "id" : "20",
  "value1" : "100",
  "value2" : "30",
  "TimestampAcquisition" : "2020-03-29 17:20:25"
}
```
The GET method, instead, returns a json message as a result of a query for a selected id. In the message only the last 10 
entries are returned:
```
{
"id": "20",
"value1":[100,100,100],
"value2":[30,30,30]
}
```

### 3.MongoDB configuration
 The MongoDB container is configured trough two main environment variables and an initialization script. The environment 
 variables are used to define the admin user and its password:
```
 MONGO_INITDB_ROOT_USERNAME: admin
 MONGO_INITDB_ROOT_PASSWORD: adminpassword
 ```
The mongo-init.js script, instead, creates a new user and a new db, according to those specified in the tornado environment 
variables:
```
db.createUser(
    {user: 'user', 
     pwd: 'userpassword', 
     roles: [
             {role: 'readWrite', 
              db:    'restappdb'
             }
            ]
    }
) 
 ```
### 4.Nginx Reverse Proxy configuration
The Nginx configuration is reported in the app.conf file:
```
upstream app_server {
    server tornado:8890;
}

server {
    listen 80;
    server_name _;
    error_log  /var/log/nginx/error.log;
    access_log /var/log/nginx/access.log;
    client_max_body_size 64M;

    location / {
    proxy_pass_header Server;
    proxy_set_header Host $http_host;
    proxy_redirect off;
    proxy_set_header X-Real-IP $remote_addr;
    proxy_set_header X-Scheme $scheme;
    proxy_pass http://app_server;
    }
}
 ```
In such a configuration the service accepts HTTP requests on the port 80 to forward them towards the app_server (tornado 
Web service application) running on the port 8890.

## References:
1. https://docs.docker.com/install/linux/docker-ce/debian/
2. https://docs.docker.com/compose/install/
